import pathpy as pp
import numpy as np
import scipy as sp
from scipy.stats import chi2
import igraph
import KOrderPathModel

def addResult(values, stream):
    line = ''
    for i in range(len(values)):
        line += str(values[i])
        if i < len(values) - 1:
            line += ','
        else:
            line += '\n'
    stream.write(line)
    stream.flush()


n_range = np.linspace(10, 1000, 20)

# Note: use separate run for fixed n and variable k to generate output for density plots
k_range = np.linspace(3, 3, 1)
order_range = np.linspace(1, 6, 6)
sample_range = np.logspace(2, 6, 30)

path_length = 15
samples = 20

num_runs = len(n_range) * len(k_range) * len(order_range) *  len(sample_range) * samples
print('Max. number of simulation runs: ' + str(num_runs))

filename = './output/output_min_sample_size.csv'
filename_detected = './output/output_detected_orders.csv'

run = 0

output = open(filename, 'w')
output.write('n,k,trueOrder,MinSamples_AIC,MinSamples_BIC,MinSamples_MOG\n')
output.flush()

output_detected = open(filename_detected, 'w')
output_detected.write('n,k,trueOrder,L,pL,detectedOrder_AIC,detectedOrder_BIC,detectedOrder_MOG\n')
output_detected.flush()

# for all node degrees
for deg in k_range:
    # for all network sizes
    for n in n_range:        

        # for all true Markov orders
        for order in order_range:           
            
            # for all samples
            for _ in range(samples):            
                
                # generate random (strongly connected) network
                g = igraph.Graph.Erdos_Renyi(n=int(n), m=int(n*deg), directed=True, loops=True).clusters(mode='STRONG').giant()

                # generate k-th-order path model
                pathModel = KOrderPathModel.KOrderPathModel(g, k=int(order))

                # zero indicates that the correct order 
                # has not yet been detected
                min_BIC = 0
                min_AIC = 0
                min_MOG = 0

                # for all sample sizes
                for pc in sample_range:
                    
                    # generate new path sample with size pc
                    if min_BIC == 0 or min_AIC == 0 or min_MOG == 0:
                        paths = pathModel.generatePaths(pathCount=int(pc), pathLength=path_length)                
                    
                    # if the BIC or AIC have not yet detected the correct order, 
                    # we fit a Markov sequence model and estimate the order
                    if min_BIC == 0 or min_AIC == 0:
                        ms = pp.MarkovSequence(paths.getSequence())

                        if min_BIC == 0:
                            detected_BIC = ms.estimateOrder(maxOrder=int(order)+1, method='BIC')
                            print('BIC: ' + str(detected_BIC) + ' vs ' + str(order))
                            if detected_BIC == int(order):
                                min_BIC = pc

                        if min_AIC == 0:
                            detected_AIC = ms.estimateOrder(maxOrder=int(order)+1, method='AIC')
                            print('AIC: ' + str(detected_AIC) + ' vs ' + str(order))
                            if detected_AIC == int(order):
                                min_AIC = pc

                    # if the multi-order model has not yet detected the correct order
                    # we fit a multi-order graphical model and estimate the order
                    if min_MOG == 0:
                        m = pp.MultiOrderModel(paths, maxOrder=int(order)+1)
                        detected_MOG = m.estimateOrder(paths, maxOrder=int(order)+1, significanceThreshold=0.01)
                        print('MOC: ' + str(detected_MOG) + ' vs ' + str(order))
                        if detected_MOG == int(order):
                            min_MOG = pc
                       
                    run +=1

                    addResult([int(n), int(deg), int(order), int(pc), int(path_length), int(detected_AIC), int(detected_BIC), int(detected_MOG)], output_detected)
                    print('Finished run ' + str(run) + '/' + str(num_runs) + ' for sample size ' + str(pc))
                    

                    # Skip remaining sample sizes if all methods have detected the correct order
                    if min_BIC > 0 and min_AIC > 0 and min_MOG > 0:
                        print('BREAKING')
                        break
                
                addResult([int(n), int(deg), int(order), int(min_AIC), int(min_BIC), int(min_MOG)], output)
        
            print('Finished for order = ', str(int(order)))
        print('Finished for n = ', str(int(n)))
    print('Finished for deg = ', str(int(deg)))

output.close()
print('All done.')